import sys
import os
import torch
import tqdm
import fire
from opendr.engine.target import BoundingBox3D
from opendr.engine.datasets import PointCloudsDatasetIterator
from opendr.perception.object_detection_3d import VoxelObjectDetection3DLearner
from opendr.perception.object_detection_3d import LabeledPointCloudsDatasetIterator
from opendr.perception.object_tracking_3d.ab3dmot.object_tracking_3d_ab3dmot_learner import (
    ObjectTracking3DAb3dmotLearner,
)
from opendr.perception.object_tracking_3d.datasets.kitti_tracking import (
    KittiTrackingDatasetIterator,
    LabeledTrackingPointCloudsDatasetIterator,
)

config_roots = {
    "pointpillars": os.path.join(
        ".",
        "src",
        "opendr",
        "perception",
        "object_detection_3d",
        "voxel_object_detection_3d",
        "second_detector",
        "configs",
        "pointpillars",
        "car",
    ),
    "tanet": os.path.join(
        ".",
        "src",
        "opendr",
        "perception",
        "object_detection_3d",
        "voxel_object_detection_3d",
        "second_detector",
        "configs",
        "tanet",
        "car",
    ),
}
temp_dir = "./run/models"
subsets_path = os.path.join(
    ".",
    "src",
    "opendr",
    "perception",
    "object_detection_3d",
    "datasets",
    "kitti_subsets",
)

dataset_tracking_path = "/data/sets/kitti_tracking"
datasets = {}

all_track_ids = [
    *["000" + str(i) for i in range(0, 10)],
    *["00" + str(i) for i in range(10, 20 + 1)],
]

for track_id in all_track_ids:
    datasets[track_id] = LabeledTrackingPointCloudsDatasetIterator(
        dataset_tracking_path + "/training/velodyne/" + track_id,
        dataset_tracking_path + "/training/label_02/" + track_id + ".txt",
        dataset_tracking_path + "/training/calib/" + track_id + ".txt",
    )


def save_detection_inference(
    model_type,
    device="cuda:0",
    name="pointpillars_car",
    samples_list=[1],
    config="xyres_16.proto",
    eval_suffix="classic",
    return_uncertainty=True,
    # format="kitti",
    format="ab3dmot",
):

    config = os.path.join(config_roots[model_type], config,)
    model_path = os.path.join(temp_dir, name)

    learner = VoxelObjectDetection3DLearner(
        model_config_path=config,
        device=device,
        checkpoint_after_iter=1000,
        return_uncertainty=return_uncertainty,
    )
    learner.load(model_path)

    for samples in samples_list:
        print("samples =", samples)

        for track_id in all_track_ids:

            dataset = datasets[track_id]

            print("Track", track_id)

            os.makedirs(
                os.path.join(
                    model_path,
                    "tracking_inference_detections" + ("" if format == "kitti" else ("_" + format)),
                    eval_suffix,
                    "samples_" + str(samples),
                ),
                exist_ok=True,
            )

            with open(
                os.path.join(
                    model_path,
                    "tracking_inference_detections" + ("" if format == "kitti" else ("_" + format)),
                    eval_suffix,
                    "samples_" + str(samples),
                    track_id + ".txt",
                ),
                "w",
            ) as f:
                for frame, (input, _) in tqdm.tqdm(
                    enumerate(dataset), total=len(dataset)
                ):
                    output = learner.infer(input, samples=samples)

                    result = "\n".join(
                        box.to_kitti_tracking_string(frame) if format == "kitti" else box.to_ab3dmot_string(frame) for box in output
                    )

                    if len(output) > 0:
                        result += "\n"

                    f.write(result)


def test_tracking(
    model_type,
    device="cuda:0",
    name="pointpillars_car",
    samples_list=[1],
    config="xyres_16.proto",
    eval_suffix="classic",
    return_uncertainty=True,
):

    config = os.path.join(config_roots[model_type], config,)
    model_path = os.path.join(temp_dir, name)

    learner = VoxelObjectDetection3DLearner(
        model_config_path=config,
        device=device,
        checkpoint_after_iter=1000,
        return_uncertainty=return_uncertainty,
    )
    learner.load(model_path)

    tracking_learner = ObjectTracking3DAb3dmotLearner()

    is_abdet = False
    is_input_det = True

    for samples in samples_list:
        print("samples =", samples)

        dataset = KittiTrackingDatasetIterator(
            os.path.join(
                model_path,
                "tracking_inference_detections",
                eval_suffix,
                "samples_" + str(samples),
            ),
            dataset_tracking_path + "/training/label_02/",
            "detection",
        )
        if is_abdet:
            dataset = KittiTrackingDatasetIterator(
                "run/models/ab3dmotdet",
                # dataset_tracking_path + "/training/label_02/",
                dataset_tracking_path + "/training/label_02/",
                "tracking",
                "detection" if is_input_det else "tracking",
            )
        # dataset = KittiTrackingDatasetIterator(
        #     # "run/models/ab3dmotdet",
        #     dataset_tracking_path + "/training/label_02/",
        #     dataset_tracking_path + "/training/label_02/",
        #     "tracking",
        #     # "tracking",
        # )

        with open(
            os.path.join(
                model_path, "tracking_eval_" + eval_suffix + "_samples_" + str(samples)
            ),
            "w",
        ) as f:

            if is_input_det:

                results = tracking_learner.eval(
                    dataset,
                    os.path.join(
                        model_path,
                        "log_tracking_eval_" + eval_suffix + "_samples_" + str(samples),
                    ),
                    verbose=True,
                )
            else:
                results = tracking_learner.dataset_eval(
                    dataset,
                    os.path.join(
                        model_path,
                        "log_tracking_eval_" + eval_suffix + "_samples_" + str(samples),
                    ),
                    verbose=True,
                )

            for key, item in results.items():
                f.write(key)
                f.write(str(item))
                f.write("\n\n")

                print(key)
                print(item)
                print("\n\n")


def test_model(
    model_type,
    device="cuda:0",
    name="pointpillars_car",
    samples_list=[1],
    config="xyres_16.proto",
    eval_suffix="classic",
    mode="test_tracking",
    return_uncertainty=True,
):

    if mode == "save_detection_inference":
        return save_detection_inference(
            model_type=model_type,
            device=device,
            name=name,
            samples_list=samples_list,
            config=config,
            eval_suffix=eval_suffix,
            return_uncertainty=return_uncertainty,
        )
    elif mode == "test_tracking":
        return test_tracking(
            model_type=model_type,
            device=device,
            name=name,
            samples_list=samples_list,
            config=config,
            eval_suffix=eval_suffix,
            return_uncertainty=return_uncertainty,
        )
    else:
        raise ValueError("Unknown mode " + str(mode))


def test_pointpillars(
    device="cuda:0",
    name="pointpillars_car",
    samples_list=[1],
    config="xyres_16.proto",
    eval_suffix="classic",
    mode="test_tracking",
):
    return test_model(
        "pointpillars",
        device=device,
        name=name,
        samples_list=samples_list,
        config=config,
        eval_suffix=eval_suffix,
        return_uncertainty=False,
        mode=mode,
    )


def test_tanet(
    device="cuda:0",
    name="tanet_car",
    samples_list=[1],
    config="xyres_16.proto",
    eval_suffix="classic",
    mode="test_tracking",
):
    return test_model(
        "tanet",
        device=device,
        name=name,
        samples_list=samples_list,
        config=config,
        eval_suffix=eval_suffix,
        return_uncertainty=False,
        mode=mode,
    )


def test_vnn_pointpillars(
    device="cuda:0",
    name="vnn_pointpillars_car",
    samples_list=[1, 2, 3, 4],
    config="vnn_xyres_16.proto",
    eval_suffix="covar",
    mode="test_tracking",
):
    return test_pointpillars(
        device=device,
        name=name,
        samples_list=samples_list,
        config=config,
        eval_suffix=eval_suffix,
    )


def test_vnna_pointpillars(
    device="cuda:0",
    name="vnna_pointpillars_car",
    samples_list=[1, 2, 3, 4],
    config="vnna_xyres_16.proto",
    eval_suffix="covar",
    mode="test_tracking",
):
    return test_pointpillars(
        device=device,
        name=name,
        samples_list=samples_list,
        config=config,
        eval_suffix=eval_suffix,
        mode=mode,
    )


def test_vnn_tanet(
    device="cuda:0",
    name="vnn_tanet_car",
    samples_list=[1, 2, 3, 4],
    config="vnn_xyres_16.proto",
    eval_suffix="covar",
    mode="test_tracking",
):
    return test_model(
        "tanet",
        device=device,
        name=name,
        samples_list=samples_list,
        config=config,
        eval_suffix=eval_suffix,
        mode=mode,
    )


def test_tapp(
    device="cuda:0",
    name="tapp_car",
    samples_list=[1],
    config="tapp_16.proto",
    eval_suffix="classic",
    mode="test_tracking",
):
    return test_model(
        "tanet",
        device=device,
        name=name,
        samples_list=samples_list,
        config=config,
        eval_suffix=eval_suffix,
        return_uncertainty=False,
        mode=mode,
    )


def test_vnn_tapp(
    device="cuda:0",
    name="mvnn_tapp_car",
    samples_list=[1, 2, 3, 4],
    config="vnn_tapp_16.proto",
    eval_suffix="covar",
    mode="test_tracking",
):
    return test_model(
        "tanet",
        device=device,
        name=name,
        samples_list=samples_list,
        config=config,
        eval_suffix=eval_suffix,
        mode=mode,
    )


def all_save_detection_inference(start=0, index=0, devices=1):
    models = [
        # add models
        # ("tanet_car", test_tanet),
        # ("tapp_car", test_tapp),
        # # ("pointpillars_car", test_pointpillars),
        *[("vnn_tanet_car_s" + str(i), test_vnn_tanet) for i in range(1, 3 + 1)],
        *[("badvnn_tanet_car_s" + str(i), test_vnn_tanet) for i in range(1, 3 + 1)],
        *[("vnn_tapp_car_s" + str(i), test_vnn_tapp) for i in range(1, 3 + 1)],
        # *[
        #     ("vnn_pointpillars_car_s" + str(i), test_vnn_pointpillars)
        #     for i in range(1, 7 + 1)
        # ],
        # *[
        #     ("vnna_pointpillars_car_s" + str(i), test_vnna_pointpillars)
        #     for i in range(1, 4 + 1)
        # ],
    ]

    i = index

    while i < len(models):

        print(":::", i, "/", len(models))

        if i < start:
            i += devices
            continue

        model_name, test_func = models[i]
        test_func(name=model_name, mode="save_detection_inference")

        i += devices

if __name__ == "__main__":
    fire.Fire()
